#!/usr/bin/python
# -*- coding: Windows-1251 -*-

'''
polsql99.py

Python MySQL AUX Proxy; version Alpha 0.1

This program allows for MySQL queries to be made from POL eScript then
passed to a MySQL server. The replies are serialized and passed back to pol
in pol native format.


Created on Feb 16, 2011
@author: AnDenixa, andenixa@gmail.com

'''

import sys
import threading
import time
import ConfigParser
import Queue
import random

import socket
import MySQLdb

from polsql.Unpacker import Unpacker
from polsql.Packer import Packer

import _mysql_exceptions

# and INI-like file parser (for storing a generic configuration) 
cfg = ConfigParser.ConfigParser()

# read the cfg file
try:
    cfg.read("polsql.cfg")
    DB_HOST = cfg.get("SQL", "DB_HOST")
    DB_USER = cfg.get("SQL", "DB_USER")
    DB_PASSWD = cfg.get("SQL", "DB_PASSWD")
    DB_NAME = cfg.get("SQL", "DB_NAME")
    
    AUX_PORT = cfg.getint("AUX", "POL_AUX_PORT")
    POL_HOST = cfg.get("AUX", "POL_HOST")
    
    MAX_SOCKETS = cfg.getint("AUX", "MAX_SOCKETS")
    
except ConfigParser.Error as ex:
        raise SystemExit, "Malformed polsql.cfg: %s" % ex    

receive_buf = ""
packets_queue = Queue.Queue()
replies_queue = Queue.Queue()
sql_queries_queue = Queue.Queue()

packer = Packer()
unpacker = Unpacker() 
         

class PacketProc(threading.Thread):
    def run(self):    
                    
        while True:
            
            packet_s = packets_queue.get()
            
            #print "received ", packet_s
            packet = unpacker.unpack( packet_s )        
            
            try:
                packet = unpacker.unpack( packet_s )
                packet["type"]
            except:
                print "Droping useles packet: %s" % packet_s
                continue    
                
            if packet["type"]=="query":
                    
                ##print "Received query request: " + packet["value"]
                pid = packet["pid"]
                
                query_uc = packet["sql"]
                
                query = ''.join([unichr(s) for s in query_uc])
                
                #query = query.encode("latin_1")            
                #print "query is", query
                
                # non-blocking requests (doesn't want the reply)                                
                
                if not dBHandler.state == DBHandler.CONNECTED:
                    replies_queue.put((pid, Exception("Not connected to DB!") ))                    
                    continue                           
                
                sql_queries_queue.put((pid, query))
                                                                 
                                                                                                                                                                                                                                          
class DBHandler(threading.Thread):
    
    (DISCONNECTED, CONNECTED) = range(2)     
    
    def __init__(self):
        threading.Thread.__init__(self)
        self.name = "Database Handler"
        self.state = DBHandler.DISCONNECTED        
        
    def connect(self):
        try:            
            print "Connecting to MySQL Server..",
            self.db=MySQLdb.connect(host = DB_HOST, user=DB_USER, passwd=DB_PASSWD, db=DB_NAME)
                        
            self.cur=self.db.cursor()
            print "Connected!"
            self.state = DBHandler.CONNECTED
            
        except MySQLdb.Error as ex:
            print "SQL Connect Error: %s" % ex
            self.state = DBHandler.DISCONNECTED        
                    
    def run(self):
        
        print "%s thread is running" % self.name
        
        while True:
            
            if self.state == DBHandler.DISCONNECTED:
                self.connect()
                if self.state == DBHandler.DISCONNECTED:
                    time.sleep(5)
                    continue
            
            sql_query = sql_queries_queue.get()
            
            pid = sql_query[0]
            query = sql_query[1]            
            
            try:                     
                ##print "trying ot execute %s" % query   
                self.cur.execute(query)
                res = self.cur.fetchall()
                replies_queue.put((pid, res))
                
            except _mysql_exceptions.ProgrammingError as ex:
                replies_queue.put((pid, Exception(ex)))     
                    
            except _mysql_exceptions.OperationalError as ex:
                ## disconnect
                if ex.args[0] == 2006:
                    self.state =DBHandler.DISCONNECTED
                    
                replies_queue.put((pid, Exception(ex)))    
                
            except _mysql_exceptions.IntegrityError as ex:
                    replies_queue.put((pid, Exception(ex)))
              
    
class StreamProcessor(threading.Thread):
    """Fetches POL packets out of TCP/IP stream"""
    
    stream_procs = []
    next_proc_idx = 0
    
    def __init__(self):
        threading.Thread.__init__(self)
        
        self.name = "StreamProcessor"        
        self.soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        
        StreamProcessor.stream_procs.append(self)        
    
    def run(self):
        
        receive_buf = ""                   
            
        while True:                         

            try:                                                               
                receive_buf+=self.soc.recv(1024)
            except:                
                try:
                    print "Connecting to POL AUX pkg .."
                    self.soc.connect((POL_HOST, AUX_PORT))
                    print "Connected..."
                except Exception as ex:
                    print "Connection refused; retry in %d secs (%s)" % (5, str(ex))
                    try:
                        self.soc.disconnect()
                    except:
                        pass                    
                    
                    self.soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)                    
                    time.sleep(5)                            
                                                                                            
            end_of_packet_pos = receive_buf.find("\r\n")
            
            if end_of_packet_pos!=-1:                                
                packets_queue.put(receive_buf[:end_of_packet_pos])
                receive_buf = ""                             
                

class PacketPusher(threading.Thread):
    """bakes and pushes the packets back to POL AUX"""
    
    def run(self):        
        print "PacketPusher is running..."
        
        while True:            
            query_reply = replies_queue.get()
            
            pid = query_reply[0]
            
            ## doesn't want the reply
            if pid == -1:                
                continue
                        
            reply = {"pid":query_reply[0], "value": query_reply[1]}
            
            reply_pkd = packer.pack(reply)
            
            soc_reply = packer.pack( [ord(s) for s in reply_pkd] )                        
            
            ##print "Concocted reply %s" % soc_reply
            ##print "sending..."                        
                        
            ## uses random socket
            try:                                
                                                
                str_proc = StreamProcessor.stream_procs[ StreamProcessor.next_proc_idx ]                                            
                StreamProcessor.next_proc_idx += 1                
                StreamProcessor.next_proc_idx %= len(StreamProcessor.stream_procs)
                
                ## uncomment for random instead of round-robbin
                #str_proc = random.choice(StreamProcessor.stream_procs)                                
                
                #print "Next idx", StreamProcessor.next_proc_idx
                
                #print "choose", str_proc 
                #print "Sending through socket", str_proc.soc 
                
                str_proc.soc.sendall("%s\r\n" % soc_reply)
                ##print "Sent reply", soc_reply
            except:
                print "Thread %s Unable to send data back to AUX" % self.name                        
            
pproc = PacketProc()
ppusher = PacketPusher()
dBHandler = DBHandler()       

print "Starting threads..."        

for i in range(MAX_SOCKETS):
    StreamProcessor().start()         

ppusher.start()
pproc.start()
dBHandler.start()

print "POLMySQL99 started..."
